/*=========================================================================

  Program:   Ionization FRont Interactive Tool (IFRIT)
  Language:  C++


Copyright (c) 2002-2006 Nick Gnedin 
All rights reserved.

This file may be distributed and/or modified under the terms of the
GNU General Public License version 2 as published by the Free Software
Foundation and appearing in the file LICENSE.GPL included in the
packaging of this file.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

=========================================================================*/


#include "iparticlegroup.h"


#include "iartifactcorrector.h"
#include "icommondatadistributors.h"
#include "icommoneventobservers.h"
#include "icontrolmodule.h"
#include "idatalimits.h"
#include "idatasubject.h"
#include "ierror.h"
#include "ieventobserver.h"
#include "ifunctionmapping.h"
#include "ihistogrammaker.h"
#include "ilookuptable.h"
#include "ipiecewisefunction.h"
#include "iparticleconnector.h"
#include "iparticledataconverter.h"
#include "iparticlegrouppipeline.h"
#include "iparticlesviewsubject.h"
#include "ipointglyph.h"
#include "ireplicatedactor.h"
#include "iviewmodule.h"
#include "iviewsubjectparallelpipeline.h"

#include <vtkCamera.h>
#include <vtkPolyData.h>
#include <vtkProperty.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>

//
//  Templates (needed for some compilers)
//
#include "iarraytemplate.h"

using namespace iParameter;


IVIEWSUBJECT_DEFINE_TYPE(iParticleGroup,ParticleGroup,pg);

IOBJECT_DEFINE_KEY(iParticleGroup,AttributeToColor,ac,OffsetInt,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AttributeSizeDirect,ad,Bool,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AttributeSizeExtraFactor,ae,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AttributeToSize,as,OffsetInt,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AttributeToConnect,atc,OffsetInt,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AttributeToSeparate,ats,OffsetInt,1);
IOBJECT_DEFINE_KEY(iParticleGroup,AutoScaled,au,Bool,1);
IOBJECT_DEFINE_KEY(iParticleGroup,StretchToColor,stc,Int,1);
IOBJECT_DEFINE_KEY(iParticleGroup,LowerLimitToColor,ltc,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,UpperLimitToColor,utc,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,StretchToSize,sts,Int,1);
IOBJECT_DEFINE_KEY(iParticleGroup,LowerLimitToSize,lts,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,UpperLimitToSize,uts,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,LineWidth,lw,Int,1);
IOBJECT_DEFINE_KEY(iParticleGroup,FixedSize,s,Float,1);
IOBJECT_DEFINE_KEY(iParticleGroup,SizeFunction,sf,Any,1);
IOBJECT_DEFINE_KEY(iParticleGroup,Type,t,Int,1);

//
//  Inherited keys
//
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_COMMON(iParticleGroup);
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_SHADING(iParticleGroup);
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_REPLICATING(iParticleGroup);
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_COLOR(iParticleGroup,1);
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_OPACITY(iParticleGroup,1);
IVIEWSUBJECT_DEFINE_INHERITED_KEYS_PALETTE(iParticleGroup,1);


namespace iParticleGroup_Private
{
	//
	//  Helper classes
	//
	class SizeFunctionEventObserver : public iEventObserver
	{

	public:

		vtkTypeMacro(SizeFunctionEventObserver,iEventObserver);
		static SizeFunctionEventObserver* New(iPiecewiseFunction *fun = 0, iShell *s = 0)
		{
			IERROR_ASSERT(fun);
			IERROR_ASSERT(s);
			return new SizeFunctionEventObserver(fun,s);
		}

		void AttachPipeline(iViewSubjectPipeline *pipe)
		{
			mPipeline = pipe;
		}

	protected:

		SizeFunctionEventObserver(iPiecewiseFunction *fun, iShell *s) : iEventObserver(s)
		{
			mFunction = fun; IERROR_ASSERT(fun);
			mPipeline = 0;

			mFunction->AddObserver(vtkCommand::ModifiedEvent,this);
		}

		virtual void ExecuteBody(vtkObject *caller, unsigned long event, void *)
		{
			if(caller==mFunction && event==vtkCommand::ModifiedEvent && mPipeline!=0) mPipeline->UpdateContents(iParticleGroupPipeline::_SizeFunction);
		}

	private:

		iPiecewiseFunction *mFunction;
		iViewSubjectPipeline *mPipeline;
	};


	class CameraEventObserver : public iEventObserver
	{

	public:

		vtkTypeMacro(CameraEventObserver,iEventObserver);
		static CameraEventObserver* New(iParticleGroup *parent = 0)
		{
			IERROR_ASSERT(parent);
			return new CameraEventObserver(parent);
		}

		void Activate(bool s)
		{
			if(s)
			{
				mReferenceSize = mParent->GetFixedSize();
				mParallelScale = mCamera->GetParallelScale();
				mDistance = mCamera->GetDistance();
			}
			else
			{
				mReferenceSize = -1.0;
			}
		}

	protected:

		CameraEventObserver(iParticleGroup *parent) : iEventObserver(parent->GetViewModule()->GetControlModule()->GetShell())
		{
			mParent = parent;

			mReferenceSize = -1.0;
			mParallelScale = mDistance = 0.0;

			mCamera = mParent->GetViewModule()->GetRenderer()->GetActiveCamera();
			mCamera->AddObserver(vtkCommand::ModifiedEvent,this);
			mCamera->Register(this);
		}

		~CameraEventObserver()
		{
			mCamera->RemoveObserver(this);
			mCamera->UnRegister(this);
		}

		virtual void ExecuteBody(vtkObject *caller, unsigned long event, void *)
		{
			if(mReferenceSize>0.0 && caller==mCamera && event==vtkCommand::ModifiedEvent)
			{
				if(mCamera->GetParallelProjection() != 0)
				{
					mParent->SetFixedSize(mReferenceSize*mCamera->GetParallelScale()/mParallelScale);
				}
				else
				{
					mParent->SetFixedSize(mReferenceSize*mCamera->GetDistance()/mDistance);
				}
			}
		}

	private:

		vtkCamera *mCamera;
		iParticleGroup *mParent;
		double mReferenceSize, mParallelScale, mDistance;
	};

	//
	//  Correct antialising artifact for opaque particles
	//
	class ArtifactCorrector : public iArtifactCorrector
	{

		IGENERICFILTER_DECLARE(ArtifactCorrector,iArtifactCorrector);

	protected:

		virtual bool HasArtifacts(vtkRenderer *ren, vtkProperty *prop)
		{
			//
			//  Activate for antialiased opaque verts
			//
			if(prop!=0 && ren->GetRenderWindow()!=0)
			{
				return (prop->GetOpacity()==1.0 && ren->GetRenderWindow()->GetPointSmoothing()!=0);
			}
			else return false;
		}
	};

	ArtifactCorrector::ArtifactCorrector(iViewSubject *vo) : iArtifactCorrector(vo)
	{
	}
}


using namespace iParticleGroup_Private;


iParticleGroup* iParticleGroup::New(iParticlesViewSubject *parent)
{
	IERROR_ASSERT(parent);
	iParticleGroup *tmp = new iParticleGroup(parent); IERROR_ASSERT(tmp);
	tmp->Configure();
	return tmp;
}


iParticleGroup::iParticleGroup(iParticlesViewSubject *parent) : iSolidViewSubject(parent->GetViewModule(),parent->GetDataType(),parent->GetSubjectName(),2,ViewSubject::Flag::SameColor|ViewSubject::Flag::SameOpacity|ViewSubject::Flag::SamePalette,0.999)
{
	mSubjectId = ViewSubject::Id::Particles;

	mParent = parent;
	mParentIndex = -1;

	mAttSize = mAttColor = -1;

	mAttSizeDirect = mAutoScaled = false;
	mAttSizeExtraFactor = 1.0;

	mType = PointGlyphType::Point;
	mSize = 1.0;
	mStretchToColor = iDataStretch::Log;
	mLimitsToColor[0] = 0.1f;
	mLimitsToColor[1] = 10.0f;

	mLineWidth = 1;
	
	mSF = iPiecewiseFunction::New(); IERROR_ASSERT(mSF);
	mSF->SetMinMax(0.0,1.0);
	mSF->MovePoint(0,0.0,1.0);
	mSF->MovePoint(1,1.0,1.0);
	
	mConnector = iParticleConnector::New(this); IERROR_ASSERT(mConnector);
	mHistogramMaker = iHistogramMaker::New(this->GetViewModule()); IERROR_ASSERT(mHistogramMaker);
	mFunction = new iFunctionMapping(mSF,mHistogramMaker,0); IERROR_ASSERT(mFunction);
	mSizeFunctionEventObserver = SizeFunctionEventObserver::New(mSF,this->GetViewModule()->GetControlModule()->GetShell()); IERROR_ASSERT(mSizeFunctionEventObserver);
	mCameraEventObserver = CameraEventObserver::New(this); IERROR_ASSERT(mCameraEventObserver);

	iAbortRenderEventObserver *obsAbortRender = this->GetViewModule()->GetAbortRenderEventObserver();
	mConnector->AddObserver(vtkCommand::ProgressEvent,obsAbortRender);

	mActors[0]->GetProperty()->SetPointSize(mSize);
	mActors[0]->GetProperty()->SetLineWidth(mLineWidth);
	mActors[1]->GetProperty()->SetPointSize(1);
	mActors[1]->GetProperty()->SetLineWidth(1);
	mActors[1]->VisibilityOff();
	mActors[1]->PickableOff();

	ArtifactCorrector *ac =	ArtifactCorrector::New(this); IERROR_ASSERT(ac);
	mActors[0]->InstallArtifactCorrector(ac);
	ac->Delete();

	this->SetColor(0,iColor(0,0,0));
}


iParticleGroup::~iParticleGroup()
{
	// must be deleted after the pipeline - do not remember why
	mSF->Register(this->Pipeline());
	mSF->Delete();

	mCameraEventObserver->Delete();
	mSizeFunctionEventObserver->Delete();
	mConnector->Delete();
	mHistogramMaker->Delete();
	delete mFunction;
}


void iParticleGroup::ConfigureBody()
{
	//
	//  Create pipeline (must be created after the object is fully created)
	//	
	this->AddMainPipeline(1);
	mActors[0]->SetInput(mConnector->GetOutput());
	mActors[1]->SetInput(this->Pipeline()->GetOutput(0));
	this->Pipeline()->SetNthInput(0,mConnector->GetOutput());

	this->SetStretchToColor(iDataStretch::Log);
	this->SetStretchToSize(iDataStretch::Log);
}


void iParticleGroup::FinishInitialization()
{
	iRequiredCast<SizeFunctionEventObserver>(INFO,mSizeFunctionEventObserver)->AttachPipeline(this->Pipeline());
}


void iParticleGroup::SetInput(vtkPolyData *input)
{
	mConnector->SetInput(input);
	mHistogramMaker->SetInput(input,mAttSize);
}


void iParticleGroup::ResetPipeline()
{
}


void iParticleGroup::SetLineWidth(int w)
{
	if(w>0 && w<100)
	{
		mLineWidth = w;
		mActors[0]->GetProperty()->SetLineWidth(mLineWidth);
		this->ClearCache();
	}
}


void iParticleGroup::SetType(int t)
{ 
	mType = t;
	this->Pipeline()->UpdateContents(iParticleGroupPipeline::_Type);

	switch (mType) 
	{
	case PointGlyphType::Point:
	case PointGlyphType::Tetrahydron: 
	case PointGlyphType::Cluster:
	case PointGlyphType::Galaxy: 
		{ 
			mShading = false; 
			break; 
		}
	default:
		{
			mShading = true; 
		}
	}

	if(mType == PointGlyphType::Point)
	{
		mActors[0]->GetProperty()->SetPointSize(mSize);
		mActors[1]->VisibilityOff();
	}
	else
	{
		mActors[0]->GetProperty()->SetPointSize(1.0); 
		mActors[1]->SetVisibility(mActors[0]->GetVisibility());
	}

	this->UpdateMaterialProperties();
	this->ClearCache();
}


void iParticleGroup::SetAttributeToColor(int a)
{ 
	if(a>=0 && a<this->GetLimits()->GetNumVars()) mAttColor = a; else mAttColor = -1;
	if(mAttColor >= 0) 
	{
		mActors[0]->SetScalarVisibility(true); 
		mActors[0]->ColorByArrayComponent(0,mAttColor);
		mActors[1]->SetScalarVisibility(true); 
		mActors[1]->ColorByArrayComponent(0,mAttColor);
		if(!this->GetSubject()->GetFixedLimits())
		{
			this->SetLowerLimitToColor(this->GetLimits()->GetMin(mAttColor));
			this->SetUpperLimitToColor(this->GetLimits()->GetMax(mAttColor));
		}
	} 
	else 
	{
		mActors[0]->SetScalarVisibility(false); 
		mActors[1]->SetScalarVisibility(false); 
	}
	this->UpdateColorLookupTable();
	this->ClearCache();
}


void iParticleGroup::SetAttributeToSize(int a)
{ 
	if(a>=0 && a<this->GetLimits()->GetNumVars())
	{
		mAttSize = a;
	}
	else mAttSize = -1;
	this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
	mFunction->AttachToLimits(this->GetLimits(),mAttSize);
	mHistogramMaker->SetInputComponent(mAttSize);
	this->ClearCache();
}


void iParticleGroup::SetAttributeSizeDirect(bool s)
{
	mAttSizeDirect = s;
	this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
	this->ClearCache();
}


void iParticleGroup::SetAttributeSizeExtraFactor(float f)
{
	if(f > 0.0)
	{
		mAttSizeExtraFactor = f;
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
		this->ClearCache();
	}
}


void iParticleGroup::SetFixedSize(float s)
{ 
	if(s>0.0 && s<1000.1)
	{
		mSize = s;
		if(mType == PointGlyphType::Point) mActors[0]->GetProperty()->SetPointSize(mSize); 
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_FixedSize);
		this->ClearCache();
	}
}


void iParticleGroup::SetAutoScaled(bool s)
{
	if(s != mAutoScaled)
	{
		mAutoScaled = s;
		iRequiredCast<CameraEventObserver>(INFO,mCameraEventObserver)->Activate(s);
		this->ClearCache();
	}
}



void iParticleGroup::SetStretchToColor(int s)
{
	mStretchToColor = s;
	this->UpdateColorLookupTable();
	this->ClearCache();
}


void iParticleGroup::SetLowerLimitToColor(float s)
{
	if(mAttColor >= 0)
	{
		this->GetLimits()->UpdateLowerLimit(s,mAttColor,mLimitsToColor[0],mLimitsToColor[1]);
		this->UpdateColorLookupTable();
		this->ClearCache();
	}
}


void iParticleGroup::SetUpperLimitToColor(float s)
{
	if(mAttColor >= 0)
	{
		this->GetLimits()->UpdateUpperLimit(s,mAttColor,mLimitsToColor[0],mLimitsToColor[1]);
		this->UpdateColorLookupTable();
		this->ClearCache();
	}
}


void iParticleGroup::UpdateColorLookupTable()
{
	if(mAttColor >= 0) 
	{
		mActors[0]->SetScalarRange(mLimitsToColor[0],mLimitsToColor[1]);
		mActors[0]->GetLookupTable()->SetStretch(mStretchToColor);
		mActors[1]->SetScalarRange(mLimitsToColor[0],mLimitsToColor[1]);
		mActors[1]->GetLookupTable()->SetStretch(mStretchToColor);
		this->ClearCache();
	}
}


void iParticleGroup::SetStretchToSize(int s)
{ 
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		this->GetLimits()->SetStretch(mAttSize,s);
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
		this->ClearCache();
	}
}


int iParticleGroup::GetStretchToSize() const
{ 
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		return this->GetLimits()->GetStretch(mAttSize);
	}
	else
	{
		return iDataStretch::Linear;
	}
}


void iParticleGroup::SetLowerLimitToSize(float s)
{
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		this->GetLimits()->SetLowerLimit(mAttSize,s);
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
		this->ClearCache();
	}
}


float iParticleGroup::GetLowerLimitToSize() const
{
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		return this->GetLimits()->GetLowerLimit(mAttSize);
	}
	else
	{
		return iMath::_FloatMin;
	}
}


void iParticleGroup::SetUpperLimitToSize(float s)
{
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		this->GetLimits()->SetUpperLimit(mAttSize,s);
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
		this->ClearCache();
	}
}


float iParticleGroup::GetUpperLimitToSize() const
{
	if(mAttSize>=0 && mAttSize<this->GetLimits()->GetNumVars())
	{
		return this->GetLimits()->GetUpperLimit(mAttSize);
	}
	else
	{
		return iMath::_FloatMax;
	}
}


void iParticleGroup::ShowBody(bool show)
{
	if(show)
	{
		this->SetStretchToColor(mStretchToColor);
		mActors[0]->VisibilityOn();
		if(mType != PointGlyphType::Point) mActors[1]->VisibilityOn();
	} 
	else 
	{
		mActors[0]->VisibilityOff();
		mActors[1]->VisibilityOff();
	}
}


float iParticleGroup::GetExtraMemorySize() const
{
	return mConnector->GetMemorySize();
}


//
//  Two functions used in saving/restoring the state and in creating new instances with - ask the parent to help
//
void iParticleGroup::SolidViewSubjectPackStateBody(iString &s) const
{
	this->PackValue(s,KeyType(),mType);
	this->PackValue(s,KeyAttributeToColor(),mAttColor);
	this->PackValue(s,KeyAttributeToSize(),mAttSize);
	this->PackValue(s,KeyFixedSize(),mSize);
	this->PackValue(s,KeyStretchToColor(),mStretchToColor);
	this->PackValue(s,KeyLowerLimitToColor(),mLimitsToColor[0]);
	this->PackValue(s,KeyUpperLimitToColor(),mLimitsToColor[1]);
	this->PackValue(s,KeyStretchToSize(),this->GetStretchToSize());
	this->PackValue(s,KeyLowerLimitToSize(),this->GetLowerLimitToSize());
	this->PackValue(s,KeyUpperLimitToSize(),this->GetUpperLimitToSize());
	this->PackValue(s,KeyAttributeSizeDirect(),mAttSizeDirect);
	this->PackValue(s,KeyAttributeSizeExtraFactor(),mAttSizeExtraFactor);
	this->PackValue(s,KeyLineWidth(),mLineWidth);
	this->PackValue(s,KeyAutoScaled(),mAutoScaled);

	this->PackValue(s,KeyAttributeToConnect(),this->GetAttributeToConnect());
	this->PackValue(s,KeyAttributeToSeparate(),this->GetAttributeToSeparate());

	this->PackValuePiecewiseFunction(s,KeySizeFunction(),mSF);
}


void iParticleGroup::SolidViewSubjectUnPackStateBody(const iString &s)
{
	int i; bool b; float f;

	if(this->UnPackValue(s,KeyType(),i)) this->SetType(i);
	if(this->UnPackValue(s,KeyAttributeToColor(),i)) this->SetAttributeToColor(i);
	if(this->UnPackValue(s,KeyAttributeToSize(),i)) this->SetAttributeToSize(i);
	if(this->UnPackValue(s,KeyFixedSize(),f)) this->SetFixedSize(f);
	if(this->UnPackValue(s,KeyStretchToColor(),i)) this->SetStretchToColor(i);
	if(this->UnPackValue(s,KeyLowerLimitToColor(),f)) this->SetLowerLimitToColor(f);
	if(this->UnPackValue(s,KeyUpperLimitToColor(),f)) this->SetUpperLimitToColor(f);
	if(this->UnPackValue(s,KeyStretchToSize(),i)) this->SetStretchToSize(i);
	if(this->UnPackValue(s,KeyLowerLimitToSize(),f)) this->SetLowerLimitToSize(f);
	if(this->UnPackValue(s,KeyUpperLimitToSize(),f)) this->SetUpperLimitToSize(f);
	if(this->UnPackValue(s,KeyAttributeSizeDirect(),b)) this->SetAttributeSizeDirect(b);
	if(this->UnPackValue(s,KeyAttributeSizeExtraFactor(),f)) this->SetAttributeSizeExtraFactor(f);
	if(this->UnPackValue(s,KeyLineWidth(),i)) this->SetLineWidth(i);
	if(this->UnPackValue(s,KeyAutoScaled(),b)) this->SetAutoScaled(b);
	if(this->UnPackValue(s,KeyAttributeToConnect(),i)) this->SetAttributeToConnect(i);
	if(this->UnPackValue(s,KeyAttributeToSeparate(),i)) this->SetAttributeToSeparate(i);
	if(this->UnPackValuePiecewiseFunction(s,KeySizeFunction(),mSF)) this->ClearCache(); // this unpacks into the appropriate function automatically.
}


void iParticleGroup::SetAttributeToConnect(int a)
{
	mConnector->SetAttributeToConnect(a);
	this->ClearCache();
}


int iParticleGroup::GetAttributeToConnect() const
{
	return mConnector->GetAttributeToConnect();
}


void iParticleGroup::SetAttributeToSeparate(int a)
{
	mConnector->SetAttributeToSeparate(a);
	this->ClearCache();
}


int iParticleGroup::GetAttributeToSeparate() const
{
	return mConnector->GetAttributeToSeparate();
}


void iParticleGroup::ShowColorBarsBody(bool)
{
}


bool iParticleGroup::CanBeShown() const
{
	return true;
}


iFunctionMapping* iParticleGroup::GetSizeFunction() const
{
	mCacheInvalid = true;
	return mFunction;
}


void iParticleGroup::ViewSubjectSyncWithData(const iDataSyncRequest &r)
{
	if(this->IsThereData())
	{
		mHistogramMaker->SetInput(mConnector->GetOutput(),-1);
	}
	else
	{
		mHistogramMaker->SetInput(0,-1);
	}

	if(mAttColor > this->GetLimits()->GetNumVars()) this->SetAttributeToColor(-1);
	if(mAttSize > this->GetLimits()->GetNumVars()) this->SetAttributeToSize(-1);
	if(r.Var()>0 && r.Var()==mAttColor)
	{
		this->GetLimits()->BlockNotifications(true);
		this->SetLowerLimitToColor(mLimitsToColor[0]);
		this->SetUpperLimitToColor(mLimitsToColor[1]);
		this->GetLimits()->BlockNotifications(false);
	}
	if(r.Var()>0 && r.Var()==mAttSize)
	{
		this->Pipeline()->UpdateContents(iParticleGroupPipeline::_AttToSize);
	}
	this->UpdateColorLookupTable();

	mFunction->AttachToLimits(this->GetLimits(),mAttSize);
}


iViewSubjectPipeline* iParticleGroup::CreatePipeline(int)
{
	return new iParticleGroupPipeline(this);
}


void iParticleGroup::ConfigureMainPipeline(iViewSubjectPipeline *p, int)
{
	iViewSubjectParallelPipeline *pp = iRequiredCast<iViewSubjectParallelPipeline>(INFO,p);

	pp->GetDataManager()->AddDistributor(new iPolyDataDistributor(pp->GetDataManager()));
	pp->GetDataManager()->AddCollector(new iPolyDataCollector(pp->GetDataManager(),0));
}

